import { ApolloQueryResult, gql } from "@apollo/client";
import { OperationVariables } from "@apollo/client/core/types";
import { camelCase, pascalCase } from "change-case";
import * as gqlBuilder from "gql-query-builder";
import Fields from "gql-query-builder/build/Fields";
import { GraphQLSchema, print } from "graphql";
import { GraphQLFieldMap, GraphQLInputFieldMap } from "graphql/type";
import { singular } from "pluralize";
import { DataProvider } from "ra-core";
import {
  CreateResult,
  DeleteManyResult,
  DeleteResult,
  GetListResult,
  GetManyReferenceResult,
  GetManyResult,
  GetOneResult,
  Identifier,
  UpdateManyResult,
  UpdateResult,
} from "react-admin";
import { createClient } from "../core/apollo/client";
import { getGraphQLSchema } from "../core/schema/util/getGraphQLSchema";
import { getTypeFields } from "../core/schema/util/getTypeFields";

const client = createClient();

export const dataProvider: DataProvider = {
  getList: (resource, { meta }): Promise<GetListResult> => {
    let query = meta?.query;
    let resultDataPath = meta?.resultDataPath;

    // call user defined query from screen
    if (query != null) {
      const selectionSetName = getSelectionSetName(query);
      console.log("perform GraphQL operation (query passed from screen): ", print(query));

      return client.query({ query }).then((result: ApolloQueryResult<any>) => {
        const data = resultDataPath
          ? result.data[selectionSetName][resultDataPath]
          : result.data[selectionSetName];

        const total = resultDataPath
          ? result.data[selectionSetName][resultDataPath].length
          : result.data[selectionSetName].length;

        return { data, total };
      });
    }

    const operation = getListOperationName(resource);
    const fields: Fields = getEntityFields(resource);

    query = gqlBuilder.query({
      operation,
      fields,
    });

    console.log("perform GraphQL operation: ", query);
    return client
      .query({
        query: gql(query.query),
      })
      .then((result) => ({
        data: result.data[operation],
        total: result.data[operation].length,
      }));
  },

  getOne: (resource, { id }): Promise<GetOneResult> => {
    const operation = getOneOperationName(resource);
    const fields: Fields = getEntityFields(resource);

    const query = gqlBuilder.query({
      operation,
      variables: { id: { value: id, type: "ID", required: true } },
      fields,
    });

    console.log("perform GraphQL operation: ", query);
    return client
      .query({
        query: gql(query.query),
        variables: query.variables,
      })
      .then((result) => ({
        data: result.data[operation],
      }));
  },

  getMany: (resource, params): Promise<GetManyResult> => {
    // TODO
    const result: GetManyResult = { data: [] };
    return Promise.resolve(result);
  },

  getManyReference: (
    resource,
    { target, id, sort, pagination, filter }
  ): Promise<GetManyReferenceResult> => {
    // TODO
    const result: GetManyReferenceResult = { data: [], pageInfo: {}, total: 0 };
    return Promise.resolve(result);
  },

  create: async (resource, { data, meta }): Promise<CreateResult> => {
    const createData: { __typename?: string } = { ...data };
    delete createData.__typename;
    const createInputType = getUpdateInputType(resource);

    if (meta?.mutation != null) {
      const mutation = meta!.mutation;
      const variables: OperationVariables = {
        input: createData,
      };
      const selectionSetName = getSelectionSetName(mutation);
      console.log(
        "perform GraphQL operation (mutation passed from screen): ",
        print(mutation),
        variables
      );

      const result = await client.mutate({ mutation, variables });
      return { data: result.data[selectionSetName] };
    }

    // update and create operation names are equals
    const operation = getUpdateOperationName(resource);
    const fields: Fields = getEntityFields(resource);

    const mutation = gqlBuilder.mutation({
      operation,
      variables: {
        input: { value: createData, type: createInputType, required: true },
      },
      fields,
    });

    console.log("perform GraphQL operation: ", mutation);
    const result = await client.mutate({
      mutation: gql(mutation.query),
      variables: mutation.variables,
    });
    return { data: result.data[operation] };
  },

  update: (resource, { id, data }): Promise<UpdateResult> => {
    const operation = getUpdateOperationName(resource);
    const fields: Fields = getEntityFields(resource);
    const updateInputType = getUpdateInputType(resource);

    const updateData: { id: Identifier; __typename?: string } = { id, ...data };
    delete updateData.__typename;
    const mutation = gqlBuilder.mutation({
      operation,
      variables: {
        input: { value: updateData, type: updateInputType, required: true },
      },
      fields,
    });

    console.log("perform GraphQL operation: ", mutation);
    return client
      .mutate({
        mutation: gql(mutation.query),
        variables: mutation.variables,
      })
      .then((result) => ({ data: result.data[operation] }));
  },

  updateMany: (resource, params): Promise<UpdateManyResult> => {
    // TODO
    let result: UpdateManyResult = {};
    return Promise.resolve(result);
  },

  delete: (resource, { id }): Promise<DeleteResult> => {
    const operation = getDeleteOperationName(resource);
    const mutation = gqlBuilder.mutation({
      operation,
      variables: { id: { value: id, type: "ID", required: true } },
    });

    console.log("perform GraphQL operation: ", mutation);
    return client
      .mutate({
        mutation: gql(mutation.query),
        variables: mutation.variables,
      })
      .then(() => ({ data: { id } }));
  },

  deleteMany: (resource, params): Promise<DeleteManyResult> => {
    // TODO
    const result: DeleteManyResult = {};
    return Promise.resolve(result);
  },
};

const getSelectionSetName = (query: Record<string, any>) => {
  return query.definitions[0].selectionSet.selections[0].name.value;
};

export const getUpdateInputType = (resource: string) => {
  if (resource.endsWith("DTO")) {
    return `${pascalCase(resource.replace("DTO", ""))}InputDTO`;
  }
  return pascalCase(`${resource}Input`);
};

export const getUpdateOperationName = (resource: string, meta?: { operation: string }) => {
  return meta?.operation ?? `update${pascalCase(singular(resource.replace("DTO", "")))}`;
};
export const getOneOperationName = (resource: string, meta?: { operation: string }) => {
  return meta?.operation ?? `${camelCase(singular(resource.replace("DTO", "")))}`;
};
export const getDeleteOperationName = (resource: string, meta?: { operation: string }) => {
  return meta?.operation ?? `delete${pascalCase(singular(resource.replace("DTO", "")))}`;
};

export const getListOperationName = (resource: string, meta?: { operation: string }) => {
  return meta?.operation ?? `${camelCase(singular(resource.replace("DTO", "")))}List`;
};

export const getEntityFields = (resource: string): String[] => {
  const schema: GraphQLSchema = getGraphQLSchema();
  const typeFields: GraphQLInputFieldMap | GraphQLFieldMap<any, any> = getTypeFields(
    resource,
    schema
  );

  return Object.keys(typeFields).filter((key: string | number) => {
    const type = typeFields[key].type;
    if (!("name" in type)) {
      return false;
    }

    const typeName: string = (type as any).name;
    return (
      typeName === "ID" || typeName === "String" || typeName === "Date" || typeName === "Number"
    );
  });
};
